// Device AM9519A
// Start am 25.3.2023
// sehr eingeschraenkte Funktionalitaet :
// keine rotierende Prioritaet (keine Ahnung wie die gebaut wird...)
// nur 1 Byte pro Interrupt
// Duerfte eine Menge LABs sein...

module AM9519A (BCLK, CERES_ON, DIN, WSEL, RSEL, CMD, IR0, IR1, IR2, IR3, IR4, IR5, INTACK,
				AM9519A_Q, VECTOR_Q, GINT);

	parameter		anzahl	= 7;	// alle Interrups aber 6 und 7 sind immer inaktiv
	
	input			BCLK;
	input			CERES_ON;
	input	 [7:0]	DIN;
	input			WSEL;
	input			RSEL;
	input			CMD;	// AA[2]
	input			IR0,IR1,IR2,IR3,IR4,IR5;
	input			INTACK;
	
	output	reg [7:0]	AM9519A_Q;
	output	reg [7:0]	VECTOR_Q;
	output			GINT;
	
	reg		 [7:0]	vecmem [0:7]	/* synthesis ramstyle = "no_rw_check , MLAB" */ ;
	reg		 [7:0]	int1_reg,int2_reg;
	reg		 [7:0]	imr,irr,isr,acr;
	reg				ce_on;
	reg		 [7:0]	cmd_reg,mode_reg;
	reg				execute;
	reg		 [5:0]	do_cmd;
	reg		 [7:0]	maske;
	reg		 [5:0]	clr_maske;
	reg				load_acr,load_imr,load_vem;
	reg		 [8:0]	vector;
	reg				interrupt;
	reg		 [1:0]	iak_pipe;
	
	wire			init;
	wire	 [7:0]	set_imr,clr_imr,set_irr,clr_irr,set_isr,clr_isr;
	wire			ld_acr,ld_imr,ld_vem,rst_acr;
	wire	 [5:1]	idis;
	wire	 [5:0]	i_order;
	wire			int_ack;
	wire	 [2:0]	addr_vmem;
	
	integer			i;

	always @(posedge BCLK) ce_on <= CERES_ON;	// fuer Reset Impuls
	assign init = CERES_ON & ~ce_on;
	
	always @(posedge BCLK) iak_pipe <= {iak_pipe[0],INTACK};	// fuer Interrupt Acknowledge
	assign int_ack = (iak_pipe == 2'b01);
		
	always @(posedge BCLK) int1_reg <= {2'b11,IR5,IR4,IR3,IR2,IR1,IR0};
	
	always @(posedge BCLK) int2_reg <= int1_reg;
	
	// +++++++++++++++++ Register ++++++++++++++++++++++
	
	always @(posedge BCLK)	// Interrupt Mask Register
		for (i=0; i<=anzahl; i=i+1)
			if (set_imr[i]) imr[i] <= 1'b1;
			  else
				if (clr_imr[i]) imr[i] <= 1'b0;
				  else
					if (ld_imr) imr[i] <= DIN[i];
	
	always @(posedge BCLK)	// Interrupt Request Register
		for (i=0; i<=anzahl; i=i+1)
			if (set_irr[i]) irr[i] <= 1'b1;
			  else
				if (clr_irr[i]) irr[i] <= 1'b0;
				
	always @(posedge BCLK)	// Interrupt Service Register
		for (i=0; i<=anzahl; i=i+1)
			if (clr_isr[i]) isr[i] <= 1'b0;
			  else
				if (set_isr[i]) isr[i] <= 1'b1;

	always @(posedge BCLK)	// Auto Clear Register
		if (rst_acr) acr <= 8'd0;
		  else
			if (ld_acr) acr <= DIN;
	
	// ++++++++++++++++++++++++++++++++++++++
	
	always @(posedge BCLK)
		if (WSEL && CMD) cmd_reg <= DIN[7:0];
		  else
			if (init) cmd_reg <= 8'd0;
	
	always @(posedge BCLK) execute <= (WSEL & CMD) | init;
	
	always @(cmd_reg)
		casex (cmd_reg)	//			irr_imr_isr
		  8'b0000_0000 : do_cmd = 6'b01_10_01;	// Reset: auto_clear_reg = 0 (acr)
		  8'b0001_xxxx : do_cmd = 6'b01_01_00;	// Clear IMR and IRR? Seltsam das IMR auf 0 gesetzt wird
		  8'b0010_xxxx : do_cmd = 6'b00_01_00;	// Clear IMR
		  8'b0011_xxxx : do_cmd = 6'b00_10_00;	// Set IMR
		  8'b0100_xxxx : do_cmd = 6'b01_00_00;	// Clear IRR
		  8'b0101_xxxx : do_cmd = 6'b10_00_00;	// Set IRR
		  8'b0110_xxxx : do_cmd = 6'b00_00_10;	// Clear highest priority ISR bit
		  8'b0111_xxxx : do_cmd = 6'b00_00_01;	// Clear ISR
		  default	   : do_cmd = 6'b00_00_00;
		endcase
		
	always @(cmd_reg)
		casex (cmd_reg[3:0])
			4'b0xxx : maske = 8'hFF;
			4'b1000 : maske = 8'h01;
			4'b1001 : maske = 8'h02;
			4'b1010 : maske = 8'h04;
			4'b1011 : maske = 8'h08;
			4'b1100 : maske = 8'h10;
			4'b1101 : maske = 8'h20;
			4'b1110 : maske = 8'h40;
			4'b1111 : maske = 8'h80;
		endcase
		
	always @(isr)
		casex (isr[5:0])
			6'bxx_xxx1 : clr_maske = 6'b00_0001;
			6'bxx_xx10 : clr_maske = 6'b00_0010;
			6'bxx_x100 : clr_maske = 6'b00_0100;
			6'bxx_1000 : clr_maske = 6'b00_1000;
			6'bx1_0000 : clr_maske = 6'b01_0000;
			6'b10_0000 : clr_maske = 6'b10_0000;
			default	   : clr_maske = 6'b00_0000;
		endcase

	assign clr_isr = execute ? (do_cmd[0] ? maske : (do_cmd[1] ? {2'd0,clr_maske} : 8'd0)) : 8'd0;
	assign set_isr = int_ack ? {2'd0,vector[8:3]} & ~acr : 8'd0;
	
	assign set_imr = (execute & do_cmd[3]) ? maske : 8'd0;
	assign clr_imr = (execute & do_cmd[2]) ? maske : 8'd0;

	assign set_irr = ((execute & do_cmd[5]) ? maske : 8'd0) | (~int1_reg & int2_reg);	// mode_reg[4] = Polaritaet von IREQ - nicht gebraucht
	assign clr_irr = ((execute & do_cmd[4]) ? maske : 8'd0) | (int_ack ? {2'd0,vector[8:3]} : 8'd0);
	
	always @(posedge BCLK) load_acr <= ~init & ((execute & (cmd_reg[7:4] == 4'hC)) | (load_acr & ~(WSEL & ~CMD)));
	always @(posedge BCLK) load_imr <= ~init & ((execute & (cmd_reg[7:4] == 4'hB)) | (load_imr & ~(WSEL & ~CMD)));
	always @(posedge BCLK) load_vem <= ~init & ((execute & (cmd_reg[7:5] == 3'd7)) | (load_vem & ~(WSEL & ~CMD)));
	
	assign ld_acr = load_acr & WSEL & ~CMD;
	assign ld_imr = load_imr & WSEL & ~CMD;
	assign ld_vem = load_vem & WSEL & ~CMD;
	assign rst_acr = execute & (cmd_reg == 8'd0);

	always @(posedge BCLK) if (ld_vem) vecmem[cmd_reg[2:0]] <= DIN;	// das Vectormemory wird auf 1 Byte beschraenkt
	
	always @(posedge BCLK)
		if (init) mode_reg[4:0] <= 5'd0;	// Ceres macht keinen Schreib-Zugriff !
		  else
			if (execute && (cmd_reg[7:5] == 3'd4)) mode_reg[4:0] <= cmd_reg[4:0];
			
	always @(posedge BCLK) if (execute && (cmd_reg[7:4] == 4'hA)) mode_reg[6:5] <= cmd_reg[6:5];

	always @(posedge BCLK)
		if (execute && (cmd_reg[7:4] == 4'hA) && (cmd_reg[1:0] == 2'd1)) mode_reg[7] <= 1'b1;	// ICU armed
		  else
			if (execute && (cmd_reg[7:4] == 4'hA) && (cmd_reg[1:0] == 2'd2)) mode_reg[7] <= 1'b0;	// ICU disarmed
			
	assign idis[1] = isr[0];
	assign idis[2] = isr[0] | isr[1];
	assign idis[3] = isr[0] | isr[1] | isr[2];
	assign idis[4] = isr[0] | isr[1] | isr[2] | isr[3];
	assign idis[5] = isr[0] | isr[1] | isr[2] | isr[3] |isr[4];
	
	assign i_order = irr[5:0] & ~({idis,1'b0} | imr[5:0]);

	always @(posedge BCLK)
		casex (i_order[4:0])
		  5'bxx_xx1 : vector = {6'h01,3'd0};
		  5'bxx_x10 : vector = {6'h02,3'd1};
		  5'bxx_100 : vector = {6'h04,3'd2};
		  5'bx1_000 : vector = {6'h08,3'd3};
		  5'b10_000 : vector = {6'h10,3'd4};
		  default   : vector = {6'h20,3'd5};
		endcase

	always @(posedge BCLK) interrupt <= |i_order & mode_reg[7];
	
	always @(posedge BCLK)
		casex ({CMD,mode_reg[6:5]})
		  3'b1xx : AM9519A_Q <= {~interrupt,1'b1,mode_reg[0],mode_reg[2],mode_reg[7],vector[2:0]};
		  3'b000 : AM9519A_Q <= isr;
		  3'b001 : AM9519A_Q <= imr;
		  3'b010 : AM9519A_Q <= irr;
		  3'b011 : AM9519A_Q <= acr;
		endcase
		
	assign GINT = ~interrupt | mode_reg[2];	// mode_reg[3] : Polaritaet von GINT - nicht gebraucht ! [2] = polled mode
	
	assign addr_vmem = mode_reg[1] ? 3'd0 : vector[2:0];	// 1 = common vector, 0 = individual vector
	always @(posedge BCLK) VECTOR_Q <= vecmem[addr_vmem];
	
endmodule
