// Device M3001
// Start am 27.3.2023
// Design nach Code Analyse vom 5.4.2023
// Stand vom 7.4.2023

module M3001 (BCLK, DIN, WSEL, RSEL, DATA, SEC_PULS, M3001_Q);
	
	input			BCLK;
	input	 [7:0]	DIN;
	input			WSEL;
	input			RSEL;
	input			DATA;	// AA[2]
	input			SEC_PULS;
	
	output	reg  [7:0]	M3001_Q;
	
	reg		 [7:0]	zeit_mem [0:15]	/* synthesis ramstyle = "no_rw_check , MLAB" */ ;
	reg		 [2:0]	addr;
	reg				old_rw;
	reg				request;
	reg				update;
	reg				flip;
	reg		 [2:0]	state;
	reg		 [4:0]	monat;
	reg				s_jahr;
	reg		 [8:0]	code;
	reg				carry;
	reg		 [7:0]	zeit_q;
	
	wire			ende_req;
	wire			last_state;
	wire	 [7:0]	m_tag;
	wire			limit;
	wire	 [3:0]	msb_bcd,lsb_bcd;
	wire	 [3:0]	rd_addr,wr_addr;
	wire			wr_zeit;
	
	always @(posedge BCLK) if (WSEL && !DATA) addr <= DIN[2:0];	// Adressregister
	
	always @(posedge BCLK) old_rw <= (WSEL | RSEL) & DATA & (addr == 3'd0);	// Ceres : von 5 nach 0
	assign ende_req = old_rw & ~WSEL & ~RSEL;
	
	always @(posedge BCLK)
		if (WSEL && !DATA && (DIN[2:0] == 3'd5)) request <= 1'b1;	// In der Ceres Start mit Register 5 Zugriff = Jahreszahl
		  else
			if (ende_req) request <= 1'b0;
			
	// update loeschen wenn Zeit geschrieben
	always @(posedge BCLK) update <= ~(WSEL & DATA) & (SEC_PULS | (update & ~last_state));
	
	always @(posedge BCLK)	// alte Zeit und neue Zeit
		if (request) flip <= flip;
		  else
			if (last_state) flip <= ~flip;
	
	// state = 1 : 1. Readzugriff => state = 6 : letzte Read => state = 7 : letzter Write + flip + update loeschen
	always @(posedge BCLK)
		if (request) state <= 3'd0;
		  else state <= state + {2'd0,update};
	
	assign last_state = (state == 3'd7) & ~request;

	always @(posedge BCLK)
		if (state == 3'd6) monat <= zeit_q[4:0];	// merken fuer naechste Sekunde = genial :-)
		  else
			if (WSEL && DATA && (addr == 3'd4)) monat <= DIN[4:0];
	
	always @(posedge BCLK)
		if (state == 3'd7) s_jahr <= (zeit_q[1:0] == 2'd0);	// Schaltjahr
		  else
			if (WSEL && DATA && (addr == 3'd5)) s_jahr <= (DIN[1:0] == 2'd0);
	
					// 					Schaltjahr !		April				Juni			September		November
	assign m_tag = (monat == 5'd2) ? {7'h14,s_jahr} : (((monat == 5'd4) | (monat == 5'd6) | (monat == 5'd9) | (monat == 5'h11)) ? 8'h30 : 8'h31);
	
	always @(state or m_tag)
		casex (state)
			 3'd4 : code = {1'b0, 8'h23};	// Stunde
			 3'd5 : code = {1'b1, m_tag};	// Tag
			 3'd6 : code = {1'b1, 8'h12};	// Monat
			 3'd7 : code = {1'b0, 8'h99};	// Jahr
		  default : code = {1'b0, 8'h59};	// Sekunde und Minute
		endcase
		
	assign limit = (zeit_q == code[7:0]);
	always @(posedge BCLK) carry <= (state == 3'd1) | (limit & carry);
	
	assign lsb_bcd = carry ? (limit ? {3'd0,code[8]} : ((zeit_q[3:0] == 4'd9) ? 4'd0 : zeit_q[3:0] + 4'd1)) : zeit_q[3:0];
	assign msb_bcd = carry ? (limit ?  4'd0          : ((zeit_q[3:0] == 4'd9) ? zeit_q[7:4] + 4'd1 : zeit_q[7:4])) : zeit_q[7:4];

	assign rd_addr = request ? {flip,addr} : { flip,(state - 3'd1)};
	assign wr_addr = request ? {flip,addr} : {~flip,(state - 3'd2)};
	assign wr_zeit = request ? WSEL & DATA :  state[2] | state[1];
	
	always @(posedge BCLK) if (wr_zeit) zeit_mem[wr_addr] <= request ? DIN : {msb_bcd,lsb_bcd};
	
	always @(posedge BCLK) zeit_q <= zeit_mem[rd_addr];
	
	always @(posedge BCLK) M3001_Q <= DATA ? zeit_q : 8'd0;	// bei Bit 0 auf ~DATA muss eine 0 stehen - sonst haengt Ceres
	
endmodule
