// Device SC2692
// Start am 20.3.2023

module SC2692 (BCLK, BRESET, DIN, CERES_ON, WSEL, RSEL, ADDR, RX_DAT, SC2692_Q, UA_INT, CERES_G);

	input			BCLK;
	input			BRESET;
	input	 [7:0]	DIN;
	input			CERES_ON;
	input			WSEL;
	input			RSEL;
	input	 [3:0]	ADDR;
	input	 [8:0]	RX_DAT;
	
	output	reg [7:0]	SC2692_Q;
	output	 [2:0]	UA_INT;
	
	output	 [7:5]	CERES_G;

	reg		 [4:3]	opcreg;	// Es koennte sein, das ein Test gemacht wird.
	reg		 [7:3]	opreg;
	reg		 [7:0]	ctur,ctlr,hold_a,imask;
	reg		[15:0]	counter;
	reg				run_flag;
	reg		 [3:0]	takt;
	reg		[13:0]	grob;
	reg				old_takt,old_grob;
	reg				ct_puls;
	reg				old_cou;
	reg				rxa_int,cou_int;
	
	wire	 [3:0]	limit;
	wire			uart_int;
	
	always @(posedge BCLK) if (WSEL && (ADDR == 4'hD)) opcreg <= DIN[4:3];
	
	always @(posedge BCLK or negedge BRESET)	// wegen Grafik ruecksetzen
		if (!BRESET) opreg <= 5'd0;
		  else
			if (WSEL && (ADDR[3:1] == 3'd7)) opreg <= ADDR[0] ? (opreg & ~DIN[7:3]): (opreg | DIN[7:3]);	//  15 = RESET, 14 = SET
			
	assign CERES_G = {~opreg[7],~opreg[6],opreg[5]};
	
	// ++++++++++++++ Clock Generator ++++++++++++++++
	
	// +++++ HIER WAR DER VOLLE SCHWACHSINN : ... : takt <= takt + 4'd1; ERGAB IN DER SIMULATION "1" UND WAHRSCHEINLICH AUCH IN DER
	// REALITAET - ES KAM KEIN INTERRUPT !
	always @(posedge BCLK) takt <= (takt == limit) ? 4'd0 : takt + 4'd1;
	
	always @(posedge BCLK) old_takt <= takt[3];
	
	// dieser Zaehler geht von 1 bis 12.288 (=12*1024)
	always @(posedge BCLK) if (old_takt && !takt[3]) grob <= (grob[13] & grob[12]) ? 14'd1 : grob + 4'd1;
	
	assign limit = (grob[0] | (grob[3:0] == 4'd0)) ? 4'd13 : 4'd12;	// 7*13+9*14 = 217 Takte @50MHz => 4,340 us ~ 16 * 1/3,6864 MHz
	
	always @(posedge BCLK) old_grob <= grob[13];
	always @(posedge BCLK) ct_puls <= ~grob[13] & old_grob;	// Takt durch 12288 => 300 Hz
	
	// ++++++++++++++ Counter/Timer ++++++++++++++++++
	
	always @(posedge BCLK) if (WSEL && (ADDR == 4'h6)) ctur <= DIN;
	always @(posedge BCLK) if (WSEL && (ADDR == 4'h7)) ctlr <= DIN;
	
	always @(posedge BCLK)
		if (RSEL && (ADDR == 4'hE)) counter <= {ctur,ctlr};	// Counter Start
		  else counter <= counter - {15'd0,run_flag & ct_puls};
		  
	always @(posedge BCLK)
		if (!CERES_ON) run_flag <= 1'b0;	// ist sauber so in Simul
		  else
			if (RSEL && (ADDR[3:1] == 3'd7)) run_flag <= ~ADDR[0];	// 14 = Start, 15 = Stop
	
	always @(posedge BCLK) old_cou <= counter[0];
	
													// Set-Impuls nur 1 Takt lang, Reset ist Stop-Command
	always @(posedge BCLK) cou_int <= run_flag & (((counter == 16'd0) & old_cou) | cou_int );
	
	// ++++++++++++++ Empfaenger Kanal A ist Keyboard ++++++++++++++++
	
	always @(posedge BCLK)
		if (RX_DAT[8]) hold_a <= RX_DAT[7:0];
	
	always @(posedge BCLK) rxa_int <= CERES_ON & (RX_DAT[8] | (rxa_int & ~(RSEL & (ADDR == 4'h3))));
	
	// +++++++++++++++ Interrupt Verarbeitung +++++++++++++
	
	always @(posedge BCLK) if (WSEL && (ADDR == 4'h5)) imask <= DIN;	// Mask Register
	
	assign uart_int = |(imask & {4'd0,cou_int,1'b0,rxa_int,1'b0});
	
	assign UA_INT = {~uart_int,~rxa_int,~cou_int};
	
	// +++++++++++++++ Output ++++++++++++++++++
	
	always @(posedge BCLK)
		casex (ADDR)
		  4'h1	  : SC2692_Q <= {7'd0,rxa_int};
		  4'h3	  : SC2692_Q <= hold_a;
		  4'h5	  : SC2692_Q <= {4'd0,cou_int,1'b0,rxa_int,1'b0};
		  4'h6	  : SC2692_Q <= counter[15:8];
		  4'h7	  : SC2692_Q <= counter[7:0];
		  default : SC2692_Q <= 8'hxx;
		endcase
	
endmodule
