// Z85C30 - sehr,sehr rudimentaer...
// Start am 2.4.2023

module Z85C30 (BCLK, CERES_ON, WSEL, RSEL, ADDR, DIN, PS2_DAT, Z85C30_Q, SCC_INT, ALARM);

	input			BCLK;
	input			CERES_ON;
	input			WSEL;
	input			RSEL;
	input	 [1:0]	ADDR;
	input	 [7:0]	DIN;
	input	 [8:0]	PS2_DAT;
	
	output	reg [7:0]	Z85C30_Q;
	output	reg		SCC_INT;
	output	reg		ALARM;
	
	reg		 [7:0]	mouse_fifo [0:15]	/* synthesis ramstyle = "no_rw_check , MLAB" */ ;
	reg				reg_sel;
	reg		 [3:0]	sel_reg;
	reg		 [7:0]	reg_3;
	reg				rx_avail;
	reg		 [1:0]	phase;
	reg		[16:0]	counter;
	reg				old_cou;
	reg		 [7:0]	in_data,rx_data;
	reg		 [1:0]	minus;
	reg		 [3:0]	wr_poi,rd_poi;
	
	wire			write;
	wire			rfifo,wfifo;
	wire			wrr3;
	wire			next;
	wire			reset;
	
	assign write = WSEL & ~ADDR[1];	// nur Kanal B
	
	always @(posedge BCLK) reg_sel <= CERES_ON & ~RSEL & (write ? (reg_sel ? 1'b0 : (DIN[2:0] != 3'd0)) : reg_sel);

	always @(posedge BCLK) if (write) sel_reg <= reg_sel ? 4'd0 : DIN[3:0];
	
	assign wrr3 = write & reg_sel & (sel_reg == 4'd3);
	
	always @(posedge BCLK)
		if (!CERES_ON) reg_3 <= 8'd0;
		  else
			if (wrr3) reg_3 <= DIN;	// RX Control, Bit 0 RX Enable
			
	always @(posedge BCLK)
		if (!reg_3[0]) phase <= 2'd3;
		  else
			if (next) phase <= 2'd0;
			  else
				if (PS2_DAT[8]) phase <= phase + {1'b0,(phase != 2'd3)};
			
	always @(posedge BCLK)	// Sperrphase ca. 2 ms
		if (phase != 2'd3) counter <= 17'd0;
		  else
			if (reset) counter <= 17'd0;
			  else
				counter <= counter + 17'd1;
				
	always @(posedge BCLK) old_cou <= counter[16];
	assign next = ~counter[16] & old_cou;	// fallende Flanke
	
	assign reset = (phase == 2'd3) & PS2_DAT[8];	// in der Zeit keinen Datenempfang
	
	always @(posedge BCLK) if (PS2_DAT[8] && (phase == 2'd0)) minus <= PS2_DAT[5:4];
	
	always @(*)
		casex (phase)	// MM-Format Ceres : 100,xs,ys,lb,mb,rb  <=>  yov,xov,ys,xs,1,mb,rb,lb	: PS2
						//					 0 x6-x0	Betrag	 <=>  x7-x0	vorzeichenbehaftet
						//					 0 y6-y0			 <=>  y7-y0
		  2'd0 : in_data = {3'b100,~PS2_DAT[4],~PS2_DAT[5],PS2_DAT[0],PS2_DAT[2],PS2_DAT[1]};
		  2'd1 : in_data = {1'b0,(minus[0] ? (7'd0 - PS2_DAT[6:0]) : PS2_DAT[6:0])};
		  2'd2 : in_data = {1'b0,(minus[1] ? (7'd0 - PS2_DAT[6:0]) : PS2_DAT[6:0])};
		  default : in_data = 8'hxx;
		endcase
	
	assign rfifo = RSEL & (ADDR == 2'd1);
	assign wfifo = PS2_DAT[8] & ~((rd_poi - wr_poi) == 4'd1) & ~(phase == 2'd3);
	
	always @(posedge BCLK) wr_poi <= reg_3[0] ? wr_poi + {3'd0,wfifo} : 4'd0;
	always @(posedge BCLK) rd_poi <= reg_3[0] ? rd_poi + {3'd0,rfifo} : 4'd0;
	
	always @(posedge BCLK) if (wfifo) mouse_fifo[wr_poi] <= in_data;
	always @(posedge BCLK) rx_data <= mouse_fifo[rd_poi];
	
	always @(posedge BCLK) rx_avail <= (wr_poi != rd_poi);
	
	always @(posedge BCLK) Z85C30_Q <= ADDR[1] ? 8'd0 : (ADDR[0] ? rx_data : ((sel_reg == 4'd0) ? {7'd0,rx_avail} : rx_data));
	
	// Die Mouse ist auf dem B-Kanal , ADDR[1] = 0 , und macht keinen Interrupt. Der ist reseviert fuer das Netzwerk = A-Kanal
	always @(posedge BCLK) SCC_INT <= 1'b1;
	
	always @(posedge BCLK) ALARM <= CERES_ON & (((rd_poi - wr_poi) == 4'd1) | ALARM);	// Errorfall
	
endmodule
