// DMA State Maschine

module DMA_SM (BCLK, MCLK, BRESET, AA, DRAM_QDMA, OROM, OROM_Q, OROM_RDY, RD_DONE, GRAF_REQ, GRAF_ADR, RDD, DMA_DQ, DMA_REQ, DMA_ADR );

	input			BCLK;
	input			MCLK;
	input			BRESET;
	input	[23:2]	AA;	// 16MB
	input	[31:0]	DRAM_QDMA;

	input			OROM;
	output	reg	[31:0]	OROM_Q;
	output	reg		OROM_RDY;
	
	input	 		RD_DONE;
	input			GRAF_REQ;
	input	[25:9]	GRAF_ADR;

	output			RDD;	// lokal
	output	[63:0]	DMA_DQ;
	
	output	reg		DMA_REQ;
	output	reg	[25:9]	DMA_ADR;
	
	reg		[31:0]	cache 		[511:0]	/* synthesis ramstyle = "no_rw_check" */ ;	// 2KB size
	reg		 [5:0]	state;
	reg				done;	// Synchronizer
	reg				d_sync;	// Synchronizer
	
	reg		 [3:0]	match;
	reg		[23:9]	adr_reg0,adr_reg1,adr_reg2,adr_reg3;
	reg		[11:0]	prio;
	reg				dma_rom;
	reg		 [1:0]	block;
	reg		 [7:0]	wr_addr;
	reg		 [1:0]	pipe_orom;
	
	reg		 [1:0]	word_count;
	reg		[31:0]	low_word;
	reg				dma_gra;
	
	wire			do_dma,dma_run,dma_mux,dma_done,do_done;
	wire	 [1:0]	sel_blk,diff;
	wire	 [2:0]	carry_0,carry_1,carry_2,carry_3;
	
	// Synchronize
	always @(negedge MCLK) d_sync	  <= RD_DONE;
	always @(posedge BCLK) done		  <= d_sync;
	
	always @(posedge BCLK or negedge BRESET)
		if (!BRESET) state <= 6'd0;
		  else
			casex ({GRAF_REQ,do_dma,done,state[2:0]})
			  6'b00x_000 : state <= 6'b000_000;	// nix zu tun
			  6'b1xx_000 : state <= 6'b000_001;	// Grafik DMA - higher Prio
			  6'b01x_000 : state <= 6'b001_101;	// ROM Read DMA - MUX umschalten
			// Grafik DMA
			  6'bxx0_001 : state <= 6'b000_001;	// warten auf RD_DONE = 1
			  6'bxx1_001 : state <= 6'b000_010; 
			  6'bxx1_010 : state <= 6'b000_010;	// warten auf RD_DONE = 0
			  6'bxx0_010 : state <= 6'b000_011;
			  6'bxxx_011 : state <= 6'b000_000;	// Leertakt - jetzt zurueck
			// ROM Read
			  6'bxx0_101 : state <= 6'b011_101;	// DMA_REQ raus - warten auf RD_DONE = 1
			  6'bxx1_101 : state <= 6'b001_110;	// DMA_REQ weg
			  6'bxx1_110 : state <= 6'b001_110;	// warten auf RD_DONE = 0
			  6'bxx0_110 : state <= 6'b100_111;
			  6'bxxx_111 : state <= 6'b000_100;	// Leertakt
			  6'bxxx_100 : state <= 6'b000_000;	// 2. Leertakt - jetzt zurueck
			  default	 : state <= 6'd0;
			endcase
			
	assign dma_run  = state[2];
	assign dma_mux	= state[3];
	assign dma_done = state[5];
	
	// bei GRAFIK DMA ein Takt zusaetzliches Delay
	always @(posedge BCLK) DMA_REQ <= dma_mux ? state[4]		: GRAF_REQ;
	always @(posedge BCLK) DMA_ADR <= dma_mux ? {2'd3,AA[23:9]} : GRAF_ADR;
	
	// ++++++++++++++++++++++++ ROM Cache 4 way associative ++++++++++++++++++++++++++++++++++++++
	
	assign diff = sel_blk[1] ? (sel_blk[0] ? prio[10:9] : prio[7:6]) : (sel_blk[0] ? prio[4:3] : prio[1:0]);
	
	assign carry_3 = {1'b0,prio[10:9]} - {1'b0,diff};
	assign carry_2 = {1'b0,prio[7:6]}  - {1'b0,diff};
	assign carry_1 = {1'b0,prio[4:3]}  - {1'b0,diff};
	assign carry_0 = {1'b0,prio[1:0]}  - {1'b0,diff};
	
	always @(posedge BCLK or negedge BRESET)
		if (!BRESET) prio <= 12'b011_010_001_000;
		  else
		    if (dma_done || do_done)
			  begin
				prio[11:9] <= (sel_blk == 2'd3) ? 3'd4 : {prio[11],(prio[10:9] + {1'b0,carry_3[2]}) };
				prio[8:6]  <= (sel_blk == 2'd2) ? 3'd4 : {prio[8], (prio[7:6]  + {1'b0,carry_2[2]}) };
				prio[5:3]  <= (sel_blk == 2'd1) ? 3'd4 : {prio[5], (prio[4:3]  + {1'b0,carry_1[2]}) };
				prio[2:0]  <= (sel_blk == 2'd0) ? 3'd4 : {prio[2], (prio[1:0]  + {1'b0,carry_0[2]}) };
			  end
			  
	always @(posedge BCLK) if (dma_done && (sel_blk == 2'd3)) adr_reg3 <= AA[23:9];
	always @(posedge BCLK) if (dma_done && (sel_blk == 2'd2)) adr_reg2 <= AA[23:9];
	always @(posedge BCLK) if (dma_done && (sel_blk == 2'd1)) adr_reg1 <= AA[23:9];
	always @(posedge BCLK) if (dma_done && (sel_blk == 2'd0)) adr_reg0 <= AA[23:9];

	always @(posedge BCLK) match[3] <= prio[11] & (adr_reg3 == AA[23:9]);
	always @(posedge BCLK) match[2] <= prio[8]  & (adr_reg2 == AA[23:9]);
	always @(posedge BCLK) match[1] <= prio[5]  & (adr_reg1 == AA[23:9]);
	always @(posedge BCLK) match[0] <= prio[2]  & (adr_reg0 == AA[23:9]);

	always @(posedge BCLK) pipe_orom <= (~OROM | OROM_RDY) ? 2'd0 : pipe_orom + {1'b0,~pipe_orom[1]};	// stops at 2
	assign do_dma  = |pipe_orom    & (match == 4'd0);
	assign do_done =  pipe_orom[0] & (match != 4'd0);
	
	assign sel_blk = dma_run ? {((prio[10:9] == 2'd3) | (prio[7:6] == 2'd3)),((prio[10:9] == 2'd3) | (prio[4:3] == 2'd3))}
							 : {(match[3] | match[2]),(match[3] | match[1])};
							 
	always @(negedge MCLK) block   <= sel_blk;
	always @(negedge MCLK) dma_rom <= dma_mux;
	
	always @(posedge MCLK) wr_addr <= (RD_DONE & dma_rom) ? wr_addr + 8'd1 : 8'd0;

	always @(posedge MCLK) if (RD_DONE && wr_addr[0]) cache[{block,wr_addr[7:1]}] <= DRAM_QDMA;
	
	always @(posedge BCLK) OROM_Q <= cache[{sel_blk,AA[8:2]}];
	always @(posedge BCLK) OROM_RDY <= OROM & ~OROM_RDY & (do_done | dma_done);
	
	// ++++++++++++++++++++++++ Grafik FIFO Interface ++++++++++++++++++++++++++
	
	always @(negedge MCLK) dma_gra <= ~dma_mux;	// Grafik

	always @(posedge MCLK) word_count <= (RD_DONE & dma_gra) ? word_count + 2'd1 : 2'd0;
	
	always @(posedge MCLK) if (word_count[0]) low_word <= DRAM_QDMA;

	assign RDD	  = (word_count == 2'b11);
	assign DMA_DQ = {DRAM_QDMA,low_word};
	
endmodule
