// Grafik Logik

module GRAFIK_SM (BCLK, GCLK, BRESET, WRITE, AA, DIN, CERES_G, RD_DONE, DMA_REQ, DMA_ADR,
				  FIFOQ, USEDW, OUTREG, RDFIFO, FIFO_RST  );

	input			BCLK;
	input			GCLK;
	input			BRESET;
	input			WRITE;
	input	 [3:2]	AA;
	input	[23:0]	DIN;
	
	input	 [7:5]	CERES_G;	// invert pixel, address, display on
	
	input			RD_DONE;
	output			DMA_REQ;
	output	reg	[25:9]	DMA_ADR;
	
	input	[63:0]	FIFOQ;
	input	 [8:0]	USEDW;
	
	output	reg	[26:0]	OUTREG;
	
	output	reg		RDFIFO;
	output			FIFO_RST;
	
	reg		[23:0]	fg_reg,bg_reg;
	reg		 [2:0]	dma,fof,ack;
	reg				en_video;
	reg		 [7:0]	state;
	reg		 [5:0]	sperr_cou;
	reg		 [1:0]	pixelart;	// always 00
	reg				dma_en;
	reg				fifo_fill;
//	reg				fifoleer;
	reg		 [7:0]	dma_blks;	// 200
	reg				dma_zero;
	
	wire			tick;
	
	always @(posedge BCLK or negedge BRESET)
		if (!BRESET) en_video <= 1'b0;
		  else
			if (WRITE && AA[3]) en_video <= DIN[0];
			
	always @(posedge BCLK) if (WRITE && (AA[3:2] == 2'd0)) bg_reg <= DIN;
	always @(posedge BCLK) if (WRITE && (AA[3:2] == 2'd1)) fg_reg <= DIN;
	
	always @(posedge BCLK) dma <= {dma[1:0],dma_en};	// einsynchronisieren
	always @(posedge BCLK) fof <= {fof[1:0],fifo_fill};
	always @(posedge BCLK) ack <= {ack[1:0],RD_DONE};

	always @(posedge BCLK)
		if (!en_video) state <= 8'd0;
		  else
			casex({dma[2],fof[2],ack[2],tick,dma_zero,state[3:0]})
			  9'b0xxxx_0000 : state <= 8'b0000_0000;
			  9'b1xxxx_0000 : state <= 8'b1000_0001;	// FIFO ruecksetzen
			  9'bxxxxx_0001 : state <= 8'b1000_0010;
			  9'bxxxxx_0010 : state <= 8'b0000_0011;
			  9'bxxxxx_0011 : state <= 8'b0000_0100;
		// Jetzt warten auf Request
		      9'bxxxx1_0100 : state <= 8'b0000_1000;	// dieser Fall bei Nur-Text
			  9'bx0xx0_0100 : state <= 8'b0000_0100;	// warten
			  9'bx1xx0_0100 : state <= 8'b0001_0101;
			  9'bxx0xx_0101 : state <= 8'b0001_0101;	// warten auf RD_DONE = 1
			  9'bxx1xx_0101 : state <= 8'b0000_0110;
			  9'bxx1xx_0110 : state <= 8'b0000_0110;	// warten auf RD_DONE = 0
			  9'bxx0xx_0110 : state <= 8'b0110_0111;
			  9'bxxx0x_0111 : state <= 8'b0100_0111;	// warten auf Sperrcounter
			  9'bxxx1x_0111 : state <= 8'b0000_1000;
		// Request abgearbeitet
			  9'bxxxx0_1000 : state <= 8'b0000_0100;	// warten auf naechsten Request
			  9'b1xxx1_1000 : state <= 8'b0000_1000;	// warten auf DMA Ende
			  9'b0xxx1_1000 : state <= 8'b0000_0000;	// Zurueck auf Anfang
			  default		: state <= 8'd0;
			endcase
			  
	assign FIFO_RST = state[7];
	assign DMA_REQ  = state[4];
			  
	always @(posedge BCLK)
		if (state[7]) DMA_ADR <= {8'h80,CERES_G[6],8'd0};	// 25..18,X,16..9
		  else
			if (state[5]) DMA_ADR <= DMA_ADR + 17'd1;
			
	always @(posedge BCLK)
		if (state[7]) dma_blks <= 8'd200;
		  else
			if (state[5]) dma_blks <= dma_blks - 8'd1;
			
	always @(posedge BCLK) dma_zero <= (dma_blks == 8'd0);
			
	always @(posedge BCLK) sperr_cou <= state[6] ? (sperr_cou + 6'd1) : 6'd0;
	
	always @(posedge BCLK) if (state[7]) pixelart <= 2'd0;
			
	assign tick = (sperr_cou == {~pixelart,4'hC});	// 12 Takte - 60 Takte warten bis zum naechsten DMA
	
	// ++++++++++++++++++++++++++ Grafik Modul +++++++++++++++++++++++++++++++
	
	reg				disp_on,inv_pix;
	reg				video_on;
	reg		 [1:0]	pmode;
	reg		[10:0]	h_counter;
	reg		 [9:0]	v_counter;
	reg				h_stop,v_stop;
	reg				req_en;
	reg				dis_en,hsync,vsync;
	reg				active_v;
	reg				load;
	reg		[79:0]	shifter;
	reg		[23:0]	bypass;
	
	wire	 [7:0]	bw_byte;
	wire			bw_bit;
	
	always @(posedge GCLK) disp_on  <= CERES_G[5];
	always @(posedge GCLK) inv_pix  <= CERES_G[7];
	always @(posedge GCLK) video_on <= en_video;
	always @(posedge GCLK) pmode 	<= pixelart;
	
	// Screen Refresh : 70 Hz
	// Horizontal : 1280 Pixels at 75 MHz = 15,2 us
	// Vertical   :  836 Lines 
	always @(posedge GCLK)
		if (!video_on) h_counter <= 11'd0;
			else h_counter <= h_stop ? 11'd0 : h_counter + 11'd1;
	
	always @(posedge GCLK) h_stop <= (h_counter == 11'd1278);	// used in clock 1279 +++++++++++++ Video setting ++++++++++
	
	always @(posedge GCLK) req_en <= active_v & ((h_counter == 11'd1) | (req_en & ~(h_counter == 11'd1025)));
	
	always @(posedge GCLK) dis_en <= active_v & ((h_counter == 11'd5) | (dis_en & ~(h_counter == 11'd1029)));
			
	always @(posedge GCLK) hsync  <=   (h_counter[10:5] == 6'b100_001)	// HSYNC generated at the end: 96 clocks wide
									 | (h_counter[10:6] == 5'b100_01 );
	
	always @(posedge GCLK)
		if (!video_on) v_counter <= 10'd0;
			else v_counter <= (h_stop & v_stop) ? 10'd0 : v_counter + {9'd0,h_stop};
	
	always @(posedge GCLK) v_stop <= (v_counter == 10'd835);	// +++++++++++++ Video setting ++++++++++
	
	always @(posedge GCLK)
		if (!video_on) active_v <= 1'b0;
			else active_v <= (h_stop & (v_counter == 10'd1)) | (active_v & ~(h_stop & (v_counter == 10'd801)));

	always @(posedge GCLK)
		if (!video_on) dma_en <= 1'b0;
			else dma_en <= (h_stop & (v_counter == 10'd0)) | (dma_en & ~(h_stop & (v_counter == 10'd801)));
			
	always @(posedge GCLK) fifo_fill = dma_en & ~(USEDW[8] | (USEDW[7:6] == 2'b11));

	always @(posedge GCLK) vsync  <= (v_counter[9:2] == 8'b11_0010_01);	// 4 lines wide starting at line 804 +++++++++++++++++

	always @(posedge GCLK)
		casex({pmode,h_counter[5:0]})
			8'b00_000010 : RDFIFO <= req_en;
			default		 : RDFIFO <= 1'b0;
		endcase
		
	always @(posedge GCLK) load <= RDFIFO;

	always @(posedge GCLK)
		casex({pmode,h_counter[2:0]})
			5'b00_100 : shifter <= load ? {16'hxxxx,FIFOQ} : {24'hxx_xxxx,shifter[63:8]};
			default	  : shifter <= shifter;
		endcase
		
	assign bw_byte = shifter[7:0];
	assign bw_bit  = bw_byte[h_counter[2:0]-3'd5] ^ inv_pix;
	
	always @(posedge GCLK) bypass <= disp_on ? (bw_bit ? fg_reg : bg_reg) : 24'd0;
		
	always @(posedge GCLK) OUTREG <= {vsync,hsync,dis_en,bypass};
	
	// ++++++++++++++ Info to CPU : this error should never occur +++++++++++++++++++++
//	always @(posedge GCLK) fifoleer <= video_on & ((RDFIFO & (USEDW == 9'd0)) | fifoleer);
	
endmodule
