// PS2 Port : erste Version vom 20. Januar 2008 : 97 Logikzellen mit 54 Registern
// One-Hot Kodierung : 87 LEs mit 57 Registern. Na ja , 10 weniger ...
// Warum so viele reine kombinatorische LCs ? Insgesamt 41 ! State-Maschine ... ?
// Braucht man die Moeglichkeit, die Outputs gezielt auf 0 zu setzen ?
// Muss man die Eingaenge lesen koennen ? Kann man machen ohne Aufwand.
// 10. Februar 2008 : Ueberarbeitung noetig : 1 Hold-Stufe und kein Empfang wenn Senden !
// 28. Oktober 2019 : Protokoll Ueberarbeitung noetig wegen alter PS2 Maus...

module PS2_SM
	( BCLK, BRESET, CARRY, DATIN, CLKIN, ACK, RD_DATA,
	  SEND, TX_DATA, PS2_DAT, DATOUT, CLKOUT );
	
	input BCLK,BRESET;
	input CARRY;	// 1/16 Takt von BCLK fuer Sampling
	input DATIN;	// PS2 Daten Eingang
	input CLKIN;	// PS2 Clock Eingang
	input ACK;		// Acknowledge fuer Interrupt
	input SEND;		// Sendekommando
	input  [7:0]	TX_DATA;
	output [15:0]	RD_DATA;
	output [8:0]	PS2_DAT;
	output 			DATOUT;	// PS2 Daten Ausgang
	output			CLKOUT;	// PS2 Clock Ausgang
		
// Fuer Empfaenger
	reg 		state_0;
	reg [2:1]	state;
	reg [2:0]	state_8bit;
//	reg [3:0]	state;
	reg [8:0]	outreg;					// eine FIFO Stufe ...
	reg [7:0]	shiftreg;
	reg [2:0]	clkshifter,datshifter;	// Eingangsfilter
	reg			clock,datum;
	reg			clock_old;
	reg			parity;
	reg			empf_done;
	reg			newdata;	// Flag nach aussen
	wire		next;	// fallende Flanke von Clock
	wire		go_rx;

// Fuer Sender :
	reg			txstate_0,txstate_1,txstate_2,txstate_3,txstate_4;
	reg [2:0]	txstate_8bit;
//	reg [3:0]	txstate;
	reg [12:0]	txtimer;
	reg [8:0]	txshifter;
//	reg	[1:0]	ps2data;
	reg			DATOUT;
	reg			RUNFLAG;
	reg			TXFLAG;
	reg	[12:0]	gapcount;
	reg			gogap;
	wire		no_rx,no_tx,go_tx,zeit,clr_run,lowd;
	
	parameter	param1 = 8'd180;	// Dauer von CLK=0
	parameter	param2 = 8'd150;	// Verzoegerung bis DATA=0
	
// Ab hier kommt der Empfaenger :
	
	always @(posedge BCLK or negedge BRESET)	// Upsampler : asynchron ! 24 MHz extern, 25 MHz intern
		if (!BRESET) clkshifter <= 3'h7;
		  else
		  begin
			if (CARRY) clkshifter <= {clkshifter[1:0],CLKIN};
		  end
		
	always @(posedge BCLK or negedge BRESET)	// Filter-Einheit
		if (!BRESET) clock <= 1'b1;
		  else
		  begin
			casex ({CARRY,clkshifter})
			  4'b1_000 : clock <= 1'b0;
			  4'b1_111 : clock <= 1'b1;
			  default  : clock <= clock;
			endcase
		  end

	always @(posedge BCLK) clock_old <= clock;
	
	assign next = ~clock & clock_old;	// Fallende Flanke erkennen
	
	always @(posedge BCLK or negedge BRESET)	// Upsampler : asynchron ! 24 MHz extern, 25 MHz intern
		if (!BRESET) datshifter <= 3'h7;
		  else
		  begin
			if (CARRY) datshifter <= {datshifter[1:0],DATIN};
		  end
		
	always @(posedge BCLK or negedge BRESET)	// Filter-Einheit
		if (!BRESET) datum <= 1'b1;
		  else
		  begin
			casex ({CARRY,datshifter})
			  4'b1_000 : datum <= 1'b0;
			  4'b1_111 : datum <= 1'b1;
			  default  : datum <= datum;
			endcase
		  end

	assign go_rx = ~datum & ~state_0 & (state == 2'b00) & no_tx;	// Empfaenger darf nicht laufen wenn Sender aktiv !

// Diese State Maschine mit One-Hot Kodierung spart ein LE im Empfaenger ! Mist ...

	always @(posedge BCLK or negedge BRESET)
		if (!BRESET) state_0 <= 1'b0;
		  else
			if (next) state_0 <= go_rx | (state_0 & (state_8bit != 3'b111));

	always @(posedge BCLK or negedge BRESET)
		if (!BRESET) state <= 2'b00;
		  else
			state <= next ? {state[1],(state_0 & (state_8bit == 3'b111))} : state;		// Fuer Parity und Stop Bit empfangen
			
	always @(posedge BCLK)
		if (!state_0) state_8bit <= 3'b000;
		  else
			state_8bit <= state_8bit + {2'b00,next};
			
//	always @(posedge BCLK or negedge BRESET)		// State Maschine (etwas uebertrieben ...)
//		if (!BRESET) state <= 4'h0;
//		  else
//		  begin
//			if (next)
//			  begin
//				casex ({datum,state})
//				  5'b1_0000 : state <= 4'b0000;
//				  5'b0_0000 : state <= 4'b1000;
//				  5'bx_1000 : state <= 4'b1001;		// 1. Bit empfangen
//				  5'bx_1001 : state <= 4'b1010;
//				  5'bx_1010 : state <= 4'b1011;
//				  5'bx_1011 : state <= 4'b1100;
//				  5'bx_1100 : state <= 4'b1101;
//				  5'bx_1101 : state <= 4'b1110;
//				  5'bx_1110 : state <= 4'b1111;
//				  5'bx_1111 : state <= 4'b0011;		// 8. Bit empfangen
//				  5'bx_0011 : state <= 4'b0001;		// Parity empfangen
//				  5'bx_0001 : state <= 4'b0000;		// Stop Bit empfangen => zurueck zum Start
//				  default   : state <= 4'b0000;
//				endcase
//			  end
//		  end
		
	always @(posedge BCLK)
		if (next && state_0) shiftreg <= {datum,shiftreg[7:1]};	// Schieberegister fuer Daten
		
	always @(posedge BCLK)	// ODD Parity : even data bits => Parity = 1
		if (next)
		  begin
			if (state_0) parity <= (state_8bit == 3'b000) ? datum : datum ^ parity;
			  else
			  begin
				if (state[1]) parity <= ~(datum ^ parity);	// Ergebnis sollte Null sein
				  else
					if (state[2]) parity <= ~datum | parity;		// Stop Bit = 1 ? Letzte Bit als Fehlerflag
			  end
		  end
		
//	always @(posedge BCLK)	// ODD Parity : even data bits => Parity = 1
//		if (next)
//		  begin
//			casex (state)
//			  4'b1000 : parity <= datum;
//			  4'b1001 : parity <= datum ^ parity;
//			  4'b101x : parity <= datum ^ parity;
//			  4'b11xx : parity <= datum ^ parity;
//			  4'b0011 : parity <= ~(datum ^ parity);	// Ergebnis sollte Null sein
//			  4'b0001 : parity <= ~datum | parity;		// Stop Bit = 1 ? Letzte Bit als Fehlerflag
//			  default : parity <= parity;
//			endcase
//		  end
		
	always @(posedge BCLK) empf_done <= next & state[2];
	
	always @(posedge BCLK or negedge BRESET)
		if (!BRESET) newdata <= 1'b0;
		  else
		  begin
//			if (next && (state == 4'b0001)) newdata <= 1'b1;	// Datum empfangen => Interrupt
			if (empf_done) newdata <= 1'b1;	// Datum empfangen => Interrupt
			  else
				if (ACK) newdata <= 1'b0;
		  end
		
	always @(posedge BCLK) if (empf_done) outreg <= {parity,shiftreg};		// Daten muessen erhalten bleiben !
		
	assign RD_DATA = {~CLKOUT,~DATOUT,RUNFLAG,TXFLAG,newdata,outreg[8],clock,datum,outreg[7:0]};
	
	assign PS2_DAT = {empf_done,shiftreg};	// zur CERES
	
// Ab hier kommt der Sender :

	always @(posedge BCLK or negedge BRESET)
		if (!BRESET) RUNFLAG <= 1'b0;
		  else
			RUNFLAG <= SEND | (RUNFLAG & ~clr_run);

//	assign go = RUNFLAG & datum & clock & (state == 4'h0);		// Empfaenger darf nicht laufen
	assign no_rx = ~state_0 & (state == 2'b00);					// Empfaenger darf nicht laufen
	assign no_tx = ~txstate_0 & ~txstate_1 & ~txstate_2 & ~txstate_3 & ~txstate_4;	// Sender darf nicht laufen
	assign go_tx = RUNFLAG & datum & clock & no_rx & no_tx;		// Empfaenger + Sender duerfen nicht laufen
	
// Die One-Hot Kodierung ergibt jetzt 84 LE in Summe. Na ja ...

	always @(posedge BCLK or negedge BRESET)
		if (!BRESET) txstate_0 <= 1'b0;
		  else
			txstate_0 <= go_tx | (txstate_0 & ~zeit);
			
	always @(posedge BCLK or negedge BRESET)
		if (!BRESET) txstate_1 <= 1'b0;
		  else
			txstate_1 <= lowd | (txstate_1 & ~next);
			 
	always @(posedge BCLK or negedge BRESET)
		if (!BRESET) txstate_2 <= 1'b0;
		  else
			if (next) txstate_2 <= txstate_1 | (txstate_2 & (txstate_8bit != 3'b111));	// Senden

	always @(posedge BCLK)
		if (!txstate_2) txstate_8bit <= 3'b000;
		  else
			txstate_8bit <= txstate_8bit + {2'b00,next};
			
	always @(posedge BCLK or negedge BRESET)
		if (!BRESET) txstate_3 <= 1'b0;
		  else
			if (next) txstate_3 <= txstate_2 & (txstate_8bit == 3'b111);	// Parity Ausgabe

	always @(posedge BCLK or negedge BRESET)
		if (!BRESET) txstate_4 <= 1'b0;
		  else
			if (next) txstate_4 <= txstate_3;	// Acknowlede Bit empfangen

//	always @(posedge BCLK or negedge BRESET)
//		if (!BRESET) txstate <= 4'h0;
//		  else
//			casex ({go,zeit,next,txstate})
//			  7'b0xx_0000 : txstate <= 4'b0000;
//			  7'b1xx_0000 : txstate <= 4'b0001;
//			  7'bx0x_0001 : txstate <= 4'b0001;		// Warten auf Ablauf der 64 us laut Datenblatt KMI
//			  7'bx1x_0001 : txstate <= 4'b0011;
//			  7'bxx0_0011 : txstate <= 4'b0011;		// Warten auf fallende Flanke
//			  7'bxx1_0011 : txstate <= 4'b1000;		// Jetzt Beginn der Ausgabe
//			  7'bxx0_1xxx : txstate <= txstate;
//			  7'bxx1_1000 : txstate <= 4'b1001;
//			  7'bxx1_1001 : txstate <= 4'b1010;
//			  7'bxx1_1010 : txstate <= 4'b1011;
//			  7'bxx1_1011 : txstate <= 4'b1100;
//			  7'bxx1_1100 : txstate <= 4'b1101;
//			  7'bxx1_1101 : txstate <= 4'b1110;
//			  7'bxx1_1110 : txstate <= 4'b1111;
//			  7'bxx1_1111 : txstate <= 4'b0111;
//			  7'bxx0_0111 : txstate <= 4'b0111;		// Parity Ausgabe
//			  7'bxx1_0111 : txstate <= 4'b0110;
//			  7'bxx0_0110 : txstate <= 4'b0110;
//			  7'bxx1_0110 : txstate <= 4'b0000;
//			  default     : txstate <= 4'b0;
//			endcase
			
//	assign clr_run = next & (txstate == 4'h6);
	assign clr_run = next & txstate_4;
	
	always @(posedge BCLK) if (clr_run) TXFLAG <= datum;	// Acknowledge vom Device empfangen
	
//	always @(posedge BCLK)
//		casex (txstate)
//		  4'b0000 : ps2data <= 2'b11;
//		  4'b0001 : ps2data <= 2'b00;
//		  4'b0011 : ps2data <= 2'b10;
//		  4'b1xxx : ps2data <= {1'b1,txshifter[0]};
//		  4'b0111 : ps2data <= {1'b1,~txshifter[8]};
//		  4'b0110 : ps2data <= 2'b11;
//		  default : ps2data <= 2'b11;
//		endcase
		
	always @(posedge BCLK) txtimer <= txstate_0 ? txtimer + 13'd1 : 13'd0;
//		if (txstate == 4'h0) txtimer <= 11'h000;
	
	assign zeit = (txtimer == {param1,5'b1_1111});	// vielfache von 32 Takte 
	
	always @(posedge BCLK) gogap <= RUNFLAG & (go_tx | (gogap & ~lowd));
		
	always @(posedge BCLK) gapcount <= RUNFLAG ? gapcount + {12'd0,gogap} : 13'd0;

	assign lowd = (gapcount == {param2,5'b1_1111});	// vielfache von 32 Takte
	
	always @(posedge BCLK)
		if (SEND) txshifter <= {1'b0,TX_DATA};
		  else
			if (next && txstate_2) txshifter <= {(txshifter[0] ^ txshifter[8]),1'b1,txshifter[7:1]};
			
//	assign DATOUT = ps2data[0];
//	assign CLKOUT = ps2data[1];

// Ausgaenge invertiert !
	assign CLKOUT = txstate_0;		// Zu Begin 64 us auf Null
	
// Paritybit invertiert raus
	always @(posedge BCLK) DATOUT <= txstate_1 | (txstate_2 & ~txshifter[0]) | (txstate_3 & txshifter[8]);
endmodule
