// New for SoC-DE1 16.10.2023
	
module DRAM_SM
	( MCLK, BCLK, MRESET, DC_ACC, DRAM_ADR, IC_ACC, IDRAM_ADR, DMA_ACC, DMA_WR, DMA_ADR, WB_USED, WRA, WB_BE, WR_REQ,
	  CKE_FF, CMD_FF, ADDR_FF, CLR_REQ, DATMUX, DQM_FF, WR_DONE, RD_DONE, IC_MDONE, IWCTRL, DC_MDONE, DWCTRL, WAMUX, WADDR,
	  FREEZE, LDEN, OE );
	
	input			MCLK;
	input			BCLK;
	input			MRESET;
	
	input			DC_ACC;
	input	[25:0]	DRAM_ADR;
	
	input	 		IC_ACC;
	input	[25:0]	IDRAM_ADR;
	
	input		 	DMA_ACC;
	input			DMA_WR;
	input	[25:9]	DMA_ADR;

	input			WB_USED;
	input	[25:4]	WRA;
	input	 [1:0]	WB_BE;
	input			WR_REQ;
	
	output	reg		CKE_FF;
	output	 [3:0]	CMD_FF;		// CS RAS CAS WE 
	output	reg	[14:0]	ADDR_FF;	// BA[1:0],ADDR[12:0]
	output	reg	 	CLR_REQ;
	output	reg		DATMUX;
	output	 [1:0]	DQM_FF;
	output			WR_DONE;
	output	reg		RD_DONE;
	output	reg		IC_MDONE;
	output	reg  [2:0]	IWCTRL;
	output	reg		DC_MDONE;
	output	reg  [2:0]	DWCTRL;
	output	reg		WAMUX;
	output	[11:2]	WADDR;
	output			FREEZE;
	output			LDEN;
	output	reg		OE;
	
	reg				synclk;
	reg		 [1:0]	lsbcou;
	reg		 [2:0]	state;	// zentrale State-Maschine
	reg		 [1:0]	test_bck;
	reg		 [3:0]	request;
	reg		[25:9]	dma_a;
	reg		 [1:0]	dma_c;
	reg		[25:0]	dc_a,ic_a;
	reg				dc_c,ic_c;
	reg		[25:0]	addr;
	reg				check_op;
	reg				top_match,bot_match;
	reg		 [2:0]	cmd_i,cmd_out;
	reg		[12:2]	cadr,col_adr;	// Bank Address is 12 & 11
	reg				wr_flag,wr_dma;
	reg		 [5:0]	col_cou,col_count;
	reg		 [2:0]	new_sta;
	reg		 [2:0]	dest_i,dest;
	reg		 [2:0]	frez_ctrl,do_frez;
	reg		[14:0]	adr_reg;
	reg				rd_op;
	reg				wr_up,do_wr;
	reg    	 [9:0]	reftimer;
	reg	    		refreq;
	reg				rw_acc;
	reg				oe_flag,oe_t2;
	reg	   	[14:0]	icounter;
	reg		 [2:0]	cmd_reg;
	reg	   	[14:0]	addr_i;
	reg				enadram;
	reg				dqm_lang,dqm_kurz;
	reg		 [2:0]	col_code;
	reg				inc_op;
	reg				ende;
	reg		 [1:0]	lsb_reg;
	reg				new_lsba;
	reg		 [1:0]	lsba_c,pipe_lsb;	// Addressbits [3:2] fuer WADDR
	reg				cap_puls;
	reg		 [2:0]	burst_c;
	reg				dc_ende,ic_ende;
	reg				pipeicmd;
	reg				do_count;
	reg		 [8:4]	upd_adr;
	
	wire			do_col_op;
	wire			dma_op;
	wire			ena;
	wire			autopre;
	wire	 [5:0]	inc_cadr;
	wire	 [8:4]	mid_adr;
	wire	 [3:2]	low_adr;
	wire			init_c;
	wire			burst_add;
	
	// ++++++++++++++ Synchronisation BCLK - MCLK ++++++++++++++++++++++++++++++
	
	always @(negedge MCLK) test_bck <= {test_bck[0],BCLK};

	always @(negedge MCLK)
		if (test_bck == 2'b01) synclk <= 1'b0;
		  else synclk <= ~synclk;

	// Kontroll-Signale fuer Test auf neue Operation. Neu am 4.11.2015 : Sperre fuer zu fruehe "Voegel" !
	always @(negedge MCLK) check_op <= (state == 3'd0) & synclk & enadram;
	
	// ++++++++++++++ alles einsynchroniseren auf MCLK +++++++++++++++++
	
	always @(negedge MCLK)
		begin
			dma_c <= {DMA_ACC,DMA_WR};
			dma_a <= DMA_ADR;
			dc_c  <= DC_ACC;
			dc_a  <= DRAM_ADR;
			ic_c  <= IC_ACC;
			ic_a  <= IDRAM_ADR;
		end
		
	always @(negedge MCLK) addr[25:9] <= dma_c[1] ? dma_a : (dc_c ? dc_a[25:9] : ic_a[25:9]);
	always @(negedge MCLK) addr[8:0]  <= dc_c ? dc_a[8:0] : ic_a[8:0];

	always @(negedge MCLK) top_match = (addr[25:9] == WRA[25:9]);	// Erkennen ob geschrieben werden muss
	always @(negedge MCLK) bot_match = (addr[8:4]  == WRA[8:4]);
	
	// ++++++++++++++ Auswertung der Requests +++++++++++++++++
	
	always @(negedge MCLK) request <= {WR_REQ,dma_c[1],dc_c,ic_c};	// nach Wichtigkeit angeordnet
	
	// Signale :
	// do_frez : steuert ob schreiben wegen vollem WB (Bit 2) oder wegen Addressmatch (Bit 1:top_match/0:bot_match)
	always @(*)
		casex ({check_op,refreq,request})
		  6'b11_xxxx : 	begin
						  cmd_out  <= 3'b001;		// Refresh : RAS/CAS/WE , active low
						  ADDR_FF  <= 15'hxxxx;		// Adresse egal !
						  col_adr  <= 10'hxxx;
						  wr_flag  <= 1'b0;			// damit keine Daten rauskommen ...
						  col_cou  <= 6'h00;		// keine Column Operation !
						  new_sta  <= 3'd5;		// 12 Takte MCLK Gesamtdauer
						  dest_i   <= 3'b000;
						  do_frez  <= 3'b000;
						end
		  6'b10_1xxx :	begin
						  cmd_out  <= 3'b011;		// Activate WRITE Zugriff
						  ADDR_FF  <= {WRA[12:11],WRA[25:13]};	// 1 kB Page , BA = MSB
						  col_adr  <= {WRA[12:4],2'b00};
						  wr_flag  <= 1'b0;  
						  col_cou  <= 6'h00;		// wird nicht gebraucht
						  new_sta  <= 3'd5;
						  dest_i   <= 3'b100;
						  do_frez  <= 3'b100;		// Acknowledge an WB
						end
		  6'b10_01xx :	begin
						  cmd_out  <= 3'b011;		// Activate DMA Zugriff
						  ADDR_FF  <= {addr[12:11],addr[25:13]};	// Achtung 1 kB Page !
						  col_adr  <= {addr[12:9],7'b0};	// 4 B Zugriff => 8 bit Col-Addr, volle col_adr fuer WB schreiben
						  wr_flag  <= DMA_WR;		// schreiben ins DRAM wenn "1"
						  col_cou  <= 6'h3F;		// 63+1 Zugriffe
						  new_sta  <= 3'd5;
						  dest_i   <= 3'b100;
						  do_frez  <= 3'b010;
						end
		  6'b10_001x :	begin
						  cmd_out  <= 3'b011;		// Activate DCACHE
						  ADDR_FF  <= {addr[12:11],addr[25:13]};
						  col_adr  <= addr[12:2];
						  wr_flag  <= 1'b0;
						  col_cou  <= 6'h01;		// 1+1 Zugriff
						  new_sta  <= 3'd5;
						  dest_i   <= 3'b010;
						  do_frez  <= 3'b011;
						end
		  6'b10_0001 :	begin
						  cmd_out  <= 3'b011;		// Activate ICACHE
						  ADDR_FF  <= {addr[12:11],addr[25:13]};
						  col_adr  <= addr[12:2];
						  wr_flag  <= 1'b0;
						  col_cou  <= 6'h01;
						  new_sta  <= 3'd5;
						  dest_i   <= 3'b001;
						  do_frez  <= 3'b011;
						end
		  default :		begin
						  cmd_out  <= cmd_reg;
						  ADDR_FF  <= adr_reg;
						  col_adr  <= 10'hxxx;
						  wr_flag  <= 1'b0;		// besser fuer Simulation
						  col_cou  <= 6'h00;
						  new_sta  <= 3'd0;
						  dest_i   <= 3'b000;	// sicher ist sicher
						  do_frez  <= 3'b000;
						end
		endcase
		
	assign CMD_FF = {(cmd_out[2:1] == 2'b11),cmd_out};	// CS erzeugen , BST wird unterdrueckt !

	// Steuersignale :
	always @(negedge MCLK) frez_ctrl <= do_frez;	// nur ein Takt lang !
	assign                 FREEZE  =    frez_ctrl[2]	// Nur ein Puls !!!
									 | (frez_ctrl[1] & top_match & WB_USED & (~frez_ctrl[0] | bot_match));
	
	// zeigt an das geschrieben werden muss => state "4" folgt nach "5"
	always @(negedge MCLK) if ((state == 3'd5) && (lsbcou == 2'd0)) wr_up <= FREEZE;
			
	always @(negedge MCLK) if (check_op) wr_dma <= wr_flag;		// nur fuer DMA-Write : schreiben ins DRAM !
	always @(negedge MCLK) if (check_op) dest 	<= dest_i;		// Steuerung der Datenpfade
	always @(negedge MCLK) if (check_op) rd_op	<= addr[1];		// Cache Enable , Cache Set wird nicht verwendet
	always @(negedge MCLK) if (check_op) rw_acc	<= cmd_out[1];	// wenn "0" dann Refresh !
	always @(negedge MCLK) if (check_op) do_wr	<= do_frez[2];
	
	assign dma_op = dest[2];	// bessere Lesbarkeit !
	
	// +++++++++++++++ Master State-Maaschine +++++++++++++++++++++++++++++++++
	
	always @(negedge MCLK or negedge MRESET)
		if (!MRESET) lsbcou <= 2'd0;
		  else lsbcou <= lsbcou + {1'b0,(state != 3'd0)};

	always @(negedge MCLK or negedge MRESET)
		if (!MRESET) state <= 3'd0;
		  else
		  begin
			if (check_op || (lsbcou == 2'd3))
				casex ({wr_up,do_wr,rw_acc,ende,state})	// Die Sequenz 5-2-0 wird fuer Refresh genutzt !
		// 0 : nix zu tun
				  7'bxx_xx_000 : state <= new_sta;
		// 5 : Master-State : von hier aus verteilt es sich ...
				  7'b1x_xx_101 : state <= 3'd4;	// Write noetig - egal warum
				  7'b0x_xx_101 : state <= 3'd2;	// allgemeiner Lesezugriff : Caches/DMA
		// 4 : 1. Takt Write ins DRAM
				  7'bxx_xx_100 : state <= 3'd3;	// zur Zeit bedingungslos
		// 3 : am Ende vom Write noch mal verteilen
				  7'bx1_xx_011 : state <= 3'd0;	// reiner Write , zurueck auf Anfang
				  7'bx0_xx_011 : state <= 3'd2;	// allgemeiner Lesezugriff : Caches/DMA
		// 2 : Daten lesen/schreiben : der DMA kann auch schreiben !
				  7'bxx_0x_010 : state <= 3'd0;	// Refresh ist nur 2 states lang
				  7'bxx_10_010 : state <= 3'd2;	// noch nicht beim letzten Burst
				  7'bxx_11_010 : state <= 3'd1;	// Zugriff beendet
		// 1 : letzter State
				  7'bxx_xx_001 : state <= 3'd0;	// zurueck auf Anfang
				  default	   : state <= 3'bxxx;	// sollte eigentlich nicht auftreten !!!
				endcase
		  end
			
	// +++++++++++++++ COL Zugriff verwalten ++++++++++++++++++++++++++++++++++
	// Page-Laenge ist nur 512 Worte : A8-A0 am DRAM
	
	always @(negedge MCLK) CLR_REQ <= (state == 3'd4) & ~lsbcou[1];	// Freigeben des Write Buffers
	
	assign do_col_op = (state != 3'd0) & (lsbcou == 2'd0) & rw_acc;	// Enable fuer Column-Operation, zeitgleich mit FREEZE
	
	always @(*)
		casex ({do_col_op,state})
		  4'b1_101 : col_code = {2'b10,~(FREEZE | wr_dma)};	// "WE" am Ausgang active low
		  4'b1_100 : col_code = 3'b100;						// bei 4 gibt es nur Write
		  4'b1_011 : col_code = do_wr ? 3'b111 : {2'b10,~wr_dma};	// bei 3 kann es zu Ende sein
		  4'b1_010 : col_code = {2'b10,~wr_dma};
		  4'b1_001 : col_code = 3'b111;
		  default  : col_code = 3'b111;
		endcase
		  
	always @(negedge MCLK) cmd_reg <= enadram ? col_code : cmd_i;

	always @(negedge MCLK) do_count <= (state == 3'd2) & (lsbcou == 2'd0);
	
	always @(negedge MCLK)	// Burst Zaehler - nur fuer autopre
		if (check_op) col_count <= col_cou;
		  else
			if (do_count) col_count <= col_count - {5'h0,~ende};
			
	always @(negedge MCLK) ende <= (col_count == 6'h00);
			
	always @(negedge MCLK) inc_op <= ~cmd_out[1];	// nach CAS	cadr+1		
	assign inc_cadr = {5'h0,(dma_op & ~(wr_up & ((state == 3'd5) | (state == 3'd4))))};
	
	always @(negedge MCLK)	// Column Address Zaehler nur fuer DMA !!!! Sonst nur Register inclusive Bank-Addresse
		if (check_op) cadr <= col_adr;
		  else
			if (inc_op) cadr <= {cadr[12:9],(cadr[8:3] + inc_cadr),cadr[2]};	// Fuer 512B Burst ...
	
	assign autopre =   ((state == 3'd2) & (col_count == 6'h01))
				     | ((state == 3'd4) & do_wr);	// fuer WRITE only
	
	always @(*)
		casex (state)
		  3'd5   : lsb_reg <= cadr[3:2];	// das Original kommt durch
		  3'd4   : lsb_reg <= 2'b10;		// WRITE Burst : "10" steht fest !
		  3'd3   : lsb_reg <= cadr[3:2];	// wichtig fuer den READ nach Address-Match !!!
		  3'd2   : lsb_reg <= dma_op ? cadr[3:2] : {~cadr[3],1'b0};	// 2. Teil READ Burst
	     default : lsb_reg <= cadr[3:2];
		endcase

	// Alles nur fuer den bloeden Fall das ein DMA kommt und vielleicht ein WB leeren noetig ist	
	always @(negedge MCLK) if (check_op) upd_adr <= WRA[8:4];	// wegen DMA !!!
	
	assign low_adr =  FREEZE ? 2'b00 : lsb_reg;	// Sonderfall FREEZE
	assign mid_adr = (FREEZE | (state == 3'd4)) ? upd_adr : cadr[8:4];
	
	// Column-Addresse oder DRAM Init
	always @(negedge MCLK) adr_reg <= enadram ? {cadr[12:11],addr_i[12:11],autopre,cadr[10:9],mid_adr,low_adr,1'b0} : addr_i;
	
	always @(negedge MCLK) if (check_op) pipe_lsb <= col_adr[3:2];
			
	// +++++++++++++++++++++ externe Steuersignale erzeugen ++++++++++++++++++++++++++++++++
	
	assign init_c = (state == 3'd2) & (lsbcou == 2'd0);	// Puls fuer Start des Schreibens in die Caches
	always @(negedge MCLK) cap_puls <= init_c;
	
	assign burst_add = (burst_c != 3'b000);
	
	always @(posedge MCLK or negedge MRESET)	// 3 Bit Zaehler : 4 Write-Pulse bei 7,5,3 und 1, steigende Flanke !!!
		if (!MRESET) burst_c <= 3'd0;
		  else
		  begin
			if (init_c) burst_c <= 3'b111;
			  else
				burst_c <= burst_c + {{3{burst_add}}};	// zaehlt runter auf 0
		  end

	always @(posedge MCLK) DWCTRL[2] <= cap_puls   & dest[1];	// Diese Pulse beschreiben die Capture Register der Caches
	always @(posedge MCLK) DWCTRL[1] <= burst_c[0] & dest[1] & rd_op;	// Burst Puls
	always @(posedge MCLK) DWCTRL[0] <= 1'b0;	// ist fuer INHIBIT gedacht. Wird z.Z. nicht benutzt
	
	always @(posedge MCLK) IWCTRL[2] <= cap_puls   & dest[0];
	always @(posedge MCLK) IWCTRL[1] <= burst_c[0] & dest[0] & rd_op;
	always @(posedge MCLK) IWCTRL[0] <= 1'b0;
	
	always @(posedge MCLK) WAMUX <= (init_c & dest[1] & rd_op) | (WAMUX & burst_add);	// Nur fuer Datencache !!!
			
	assign LDEN = burst_c[0];	// Lade Ausgaberegister LSW fuer READ

	always @(posedge MCLK) new_lsba <= burst_c[0];
	
	always @(posedge MCLK)	// Adress--Sequenz fuer den wichtigsten Fall : 1-0-2-3 !
		if (init_c) lsba_c <= pipe_lsb;
		  else
			if (new_lsba) lsba_c <= burst_c[1] ? {lsba_c[1],~lsba_c[0]} : {~lsba_c[1],1'b0};
			
	assign WADDR = {cadr[11:4],lsba_c};	// Addresse an die Caches -- Timing beachten ! Multicycle ... ?
			
	always @(negedge MCLK) dc_ende  <= dest[1] & (state == 3'd2) & (lsbcou == 2'd2);
	always @(posedge BCLK) DC_MDONE <= dc_ende;	// Datencache ein Takt frueher moeglich als ICache !
	always @(negedge MCLK) ic_ende  <= dest[0] & (state == 3'd2) & (lsbcou == 2'd2);
	always @(posedge BCLK) pipeicmd <= ic_ende;
	always @(posedge BCLK) IC_MDONE <= pipeicmd;
	
	// Steuerung der Datenauswahl beim Schreiben, Datenmultiplexer fuer WRITE -- Timing beachten ! Multicycle !
	always @(negedge MCLK) if (do_col_op) DATMUX <= ((state == 3'd5) & FREEZE) | (state == 3'd4);

	// WR_DONE (knappes Timing !) und RD_DONE sind einhuellende Signale da der MEMBLK einen eigenen Address-Zaehler hat !
	always @(posedge MCLK) RD_DONE <= (init_c & dma_op & ~wr_dma) | (RD_DONE & burst_add);
	assign WR_DONE = (state != 3'd0) & (state != 3'd4) & wr_dma;	// wenn Adressmatch nochmal auf 0 in state "4"
			
	// DQM Steuerung : BE beim schreiben, "0" beim lesen. Sonderfall DMA Write !
	always @(negedge MCLK) dqm_kurz <= ((state == 3'd5) | (state == 3'd3)) & (lsbcou == 2'b00) & rw_acc;	// nur einen Takt lang
	// dqm_lang ist fuer alle READ Faelle und DMA Write
	always @(negedge MCLK) if (lsbcou == 2'd1) dqm_lang <= ((state == 3'd5) | (state == 3'd3) | (state == 3'd2)) & rw_acc;
	assign DQM_FF = DATMUX ? ~WB_BE : ( (dqm_lang | dqm_kurz) ? 2'b00 : 2'b11 );

	// OE Steuerung der Datenbus-Buffer vom FPGA :
	always @(negedge MCLK) if (do_col_op) oe_flag <=  (((state == 3'd5) & FREEZE) | (state == 3'd4))	// bei WRITE OE aktiv !
													| (((state == 3'd5) | (state == 3'd3) | (state == 3'd2)) & wr_dma);
	always @(negedge MCLK) oe_t2 <= oe_flag;
	always @(posedge MCLK) OE <= oe_flag | oe_t2;	// 9 MCLKs lang
	
	// +++++++++++++++ Refresh Controller : keine High & Low Prioritaet +++++++++++++++++++++++
	
	always @(negedge MCLK)		// Refresh Timer : Periode = 7,81 us
	  if (!enadram) reftimer <= 10'h0;
		else
		begin
		  if (reftimer == 10'd780) reftimer <= 10'h0;
			else
			  reftimer <= reftimer + 10'h1;
		end
		
	always @(negedge MCLK)
	  if (!enadram) refreq <= 1'b0;
		else
		begin
		  if (reftimer == 10'd780) refreq <= 1'b1;
			else
			  if (check_op && (CMD_FF[2:1] == 2'b00)) refreq <= 1'b0;
		end
		
	// ++++++++++++++ DRAM Initialisierungs Controller +++++++++++++++++++++++++++
	// 200 us Delay @ 133,3 MHz = 26666 Clocks, CKE und DQM = High waehrend dieser Zeit
	// 15'h7000 = 28672 Takte
	
	// CKE ist aktiv high : Clock-Enable , bei 0 = Powerdown/SRF
	always @(negedge MCLK) CKE_FF <= MRESET;
	
	always @(negedge MCLK) icounter <= MRESET ? (icounter + 15'd1) : 15'h0;
	
	assign ena = (icounter[14:8] == 7'h70) & (icounter[3:0] == 4'hF) & ~enadram;	// Kommandos in 16 Takten Abstand
	
	always @(negedge MCLK)
		casex ({ena,icounter[7:4]})
		  5'b1_0001 : cmd_i <= 3'b010;	// PRE Precharge , active low !
		  5'b1_01xx : cmd_i <= 3'b001;	// 4 * ARF
		  5'b1_10xx : cmd_i <= 3'b001;	// 4 * ARF
		  5'b1_1101 : cmd_i <= 3'b000;	// MRS
		  default   : cmd_i <= 3'b111;	// NOP bzw. DESELECT
		endcase
		
	// MRS : CL=2 , Sequentiell , Burstlaenge = 4								DESELECT, PRE sind PREA !
	always @(negedge MCLK) addr_i <= ((icounter[7:4] == 4'hD) & ena) ? 15'h0022 : 15'h7FFF;
		
	// Am Ende der Init-Sequenz CMD/ADDR-MUXe umschalten
	always @(negedge MCLK) enadram <= MRESET & (enadram | (icounter[14:8] == 7'h71));
	
endmodule
