// 16.10.2023 WB updated for 64 MB

module WB_CONTROL ( BCLK, MCLK, BRESET, MRESET, CLR_REQ, DRAM_WR, FREEZE, ENBYTE, WR_ADDR,
					WR_REQ, WRA_WB, RDA_WB, WB_BE, WRA, WB_USED, ENWR );

	input			BCLK;				
	input			MCLK;
	input			BRESET;
	input			MRESET;
	input			CLR_REQ;	// kommt vom DRAM Controller
	input			DRAM_WR;
	input			FREEZE;		// das ist der Acknowledge vom DRAM Controller , das geschrieben wird !
	input	 [3:0]	ENBYTE;
	input	[25:2]	WR_ADDR;	// beschraenkt auf 64 MB

	output	reg		WR_REQ;		// an DRAM Controller
	output	 [5:2]	WRA_WB;
	output	 [5:1]	RDA_WB;
	output	 [1:0]	WB_BE;		// BE fuer das DRAM
	output	[25:4]	WRA;		// Addresse zum vergleichen mit anderen Quellen
	output	reg		WB_USED;	// WB enthaelt Daten
	output			ENWR;		// Info an CPU
	
	reg		[25:4]	WRA;
	reg				select;
	reg				used_0,used_1;
	reg		[25:4]	addresse_0,addresse_1;
	reg		[15:0]	init_be,be_feld0,be_feld1;
	reg				rd_msb;
	reg		 [2:0]	rd_lsb;
	reg		 [1:0]	sh_used_0,sh_used_1;
	reg				sel_adr;
	reg		[15:0]	shifter;
	
	wire			clr_0,clr_1;
	wire			do_flip;
	wire			update_0,update_1;
	wire			edge_0,edge_1;
	wire			run_lsb;

	// "select" waehlt den Teil vom WB aus der Daten aufnimmt. Dazu kommt used_0/1 als Indikator ob Daten drin sind.
	always @(posedge BCLK or negedge BRESET)
		if (!BRESET) select <= 1'b0;
		  else
			if (DRAM_WR && do_flip) select <= ~select;
			
	assign do_flip = ((select ? addresse_1 : addresse_0) != WR_ADDR[25:4]) & (used_0 | used_1);	// 32 MB
	
	assign ENWR =  ~(used_0 & used_1) & ~(DRAM_WR & do_flip);	// ENABLE WRITE fuer CPU
							   
	// Schreibadresse vom Write Buffer RAM
	assign WRA_WB  = {1'b0,(do_flip ? ~select : select),WR_ADDR[3:2]};	// RAM hat minimum 16 Worte ...
	
	assign update_0 = DRAM_WR & (do_flip ?  select : ~select);
	assign update_1 = DRAM_WR & (do_flip ? ~select :  select);
	
	always @(posedge BCLK or negedge BRESET)
		if (!BRESET) used_0 <= 1'b0;
			else
			begin
				if (clr_0) used_0 <= 1'b0;
				  else
					if (update_0) used_0 <= 1'b1;
			end
		
	always @(posedge BCLK or negedge BRESET)
		if (!BRESET) used_1 <= 1'b0;
			else
			begin
				if (clr_1) used_1 <= 1'b0;
				  else
					if (update_1) used_1 <= 1'b1;
			end
		
	// Die Addressen werden immer wieder beschrieben ...
	always @(posedge BCLK) if (update_0) addresse_0 <= WR_ADDR[25:4];
	always @(posedge BCLK) if (update_1) addresse_1 <= WR_ADDR[25:4];
	
	always @(WR_ADDR or ENBYTE)
		case(WR_ADDR[3:2])
			2'b00 : init_be = {12'd0,ENBYTE};
			2'b01 : init_be = { 8'd0,ENBYTE, 4'd0};
			2'b10 : init_be = { 4'd0,ENBYTE, 8'd0};
			2'b11 : init_be = {      ENBYTE,12'd0};
		endcase
		
	always @(posedge BCLK) if (update_0) be_feld0 <= used_0 ? (be_feld0 | init_be) : init_be;
	always @(posedge BCLK) if (update_1) be_feld1 <= used_1 ? (be_feld1 | init_be) : init_be;
	
	// ++++++++++++++++++++++++++++++ BCLK-MCLK Interface ++++++++++++++++++++++++++++++

	// Der FREEZE kommt so, das ein WR_REQ durch ein Address-Match geloescht werden kann : im gleichen Takt kommt
	// ein WRITE-Request vom DCACHE und ein READ-Request vom ICACHE : self-modifying-code Situation !
			
	always @(negedge MCLK) WR_REQ <= used_0 & used_1;
	
	always @(negedge MCLK) sh_used_0 <= {used_0,sh_used_0[1]};
	always @(negedge MCLK) sh_used_1 <= {used_1,sh_used_1[1]};
	
	assign edge_0 = ~sh_used_0[1] & sh_used_0[0];	// fallende Flanke
	assign edge_1 = ~sh_used_1[1] & sh_used_1[0];
	
	// Schalter fuer die Vergleichsaddresse im DRAM Controller, "select" koennte schon anders stehen !!!
	// used_0 : _____-----______------
	// used_1 : _______------------___
	// Edge               V        V
	// sel_adr: __________---------___
	always @(negedge MCLK or negedge MRESET)
		if (!MRESET) sel_adr <= 1'b0;
		  else
			if (edge_0 || edge_1) sel_adr <= (edge_0 & sh_used_1[1]) | (edge_1 & ~sh_used_0[1]);

	// Rueckmeldung an "used_*"
	assign clr_0 = CLR_REQ & ~sel_adr;
	assign clr_1 = CLR_REQ &  sel_adr;
	
	// ++++++++++++++++++++++++++++++ WB Leseaddresse ++++++++++++++++++++++++++++++++++
	// Das Lesen der Daten aus dem RAM beginnt mit dem FREEZE : deshalb rd_msb stabil halten waehrend des Lesens !
	
	always @(negedge MCLK) if (rd_lsb == 3'd0) rd_msb <= sel_adr;
	
	assign run_lsb = FREEZE | (rd_lsb != 3'd0);
	
	always @(negedge MCLK or negedge MRESET)	// der MRESET ist dem ModelSim geschuldet ...
		if (!MRESET) rd_lsb <= 3'b000;
			else rd_lsb <= rd_lsb + {2'b00,run_lsb};	// stoppt bei 0
	
	assign RDA_WB = {1'b0,rd_msb,rd_lsb};	// RAM hat minimum 16 Worte ...

	// ++++++++++++++++++++++++++++++ Signale zum DRAM Controller ++++++++++++++++++++
	
	always @(negedge MCLK) WRA     <= sel_adr ? addresse_1 : addresse_0;	// Column Teil nur zum vergleichen, Row auch fuer RAS
	always @(negedge MCLK) WB_USED <= sel_adr ? used_1 : used_0;	// Wichtig : ohne das ist ein Match nicht relevant
	
	always @(negedge MCLK)
		if (FREEZE) shifter <= sel_adr ? be_feld1 : be_feld0;
			else shifter <= {shifter[1:0],shifter[15:2]};
			
	assign WB_BE = shifter[1:0];	// erst DRAM Controller dann zum DRAM
	
endmodule
